﻿using System;
using System.IO;

namespace DIMENSION4_OTW_FORMAT
{
    // An OTW client location message.
    public class Message
    {
        const ushort MAGIC_01 = 0x98E2;
        const ushort MAGIC_02 = 0x6B02;

        public ushort Protocol_ID ;    // HEX E298 magic number
        public ushort Message_ID ;     // HEX 026B magic number
        public  ulong Tag_ID ;         // The ID on the tag label
        public   uint Flags ;          // 0 if not a valid location reading
        public  float X ;              // IEEE float X position
        public  float Y ;              // IEEE float Y position
        public  float Z ;              // IEEE float Z position
        public  float Variance ;       // IEEE float variance
        public  ulong Timestamp ;      // Timestamp (nanoseconds)

        // Note that DIMENSION4 sensors do not use NTP so the absolute
        // value of the timestamp is not correct, but the difference 
        // between two timestamps will give an accurate measure of the
        // time between the two measurements.

        // Decode a byte array containing an OTW protocol location message.
        public void Decode(byte[] bytes)
        {
            using (MemoryStream ms = new MemoryStream())
            {
                ms.Write(bytes, 0, bytes.Length);
                ms.Flush();
                ms.Position = 0;

                using (BinaryReader reader = new BinaryReader(ms))
                {
                    Protocol_ID = reader.ReadUInt16();
                    if (Protocol_ID != MAGIC_01)
                        throw new IOException(String.Format("Received invalid Protocol ID {0:X}", Protocol_ID));
                    Message_ID = reader.ReadUInt16();
                    if (Message_ID != MAGIC_02)
                        throw new IOException(String.Format("Received invalid Message ID {0:X}", Message_ID));
                    Tag_ID = SwapBytes(reader.ReadUInt64());
                    uint local_flags = SwapBytes(reader.ReadUInt32());
                    X = reader.ReadSingle();
                    Y = reader.ReadSingle();
                    Z = reader.ReadSingle();
                    Variance = reader.ReadSingle();
                    Timestamp = SwapBytes(reader.ReadUInt64());
                    Flags = local_flags;
                    
                }
            }
        }

        // Encode a byte array containing this location message
        // (which may or may not be valid).
        public byte[] Encode()
        {
            using (MemoryStream ms = new MemoryStream())
            {
                using (BinaryWriter writer = new BinaryWriter(ms))
                {
                    writer.Write(Protocol_ID);
                    writer.Write(Message_ID);
                    writer.Write(SwapBytes(Tag_ID));
                    writer.Write(SwapBytes(Flags));
                    writer.Write(X);
                    writer.Write(Y);
                    writer.Write(Z);
                    writer.Write(Variance);
                    writer.Write(SwapBytes(Timestamp));
                    return ms.ToArray();
                }
            }
        }

        // Print an OTW protocol location message.
        public void Print (bool including_invalid = false)
        {
            if (Flags > 0)
                Console.WriteLine("Time {0:X}: Tag {1:X} @ ({2:0.000},{3:0.000},{4:0.000}) variance {5:0.000}",
                    Timestamp,
                    Tag_ID,
                    X,
                    Y,
                    Z,
                    Variance);
            else if (including_invalid)
                Console.WriteLine("Time{0:X}: Tag {1:X} location invalid",
                    Timestamp,
                    Tag_ID);
        }

        // The size of a protocol message in bytes.
        public int Size
        {
            get { return 40; }
        }

        // Create a (zeroed) location message.
        public Message()
        {
            IsLittleEndian = BitConverter.IsLittleEndian;
            Protocol_ID = 0;
            Message_ID = 0;
            Tag_ID = 0;
            Flags = 0;
            X = 0.0f;
            Y = 0.0f;
            Z = 0.0f;
            Variance = 0.0f;
            Timestamp = 0;
        }

        // Create a valid location message.
        public Message (ulong tag_id,float x,float y,float z,float variance)
        {
            IsLittleEndian = BitConverter.IsLittleEndian;
            Protocol_ID = MAGIC_01;
            Message_ID = MAGIC_02;
            Tag_ID = tag_id;
            Flags = 1;
            X = x;
            Y = y;
            Z = z;
            Variance = variance;
            Timestamp = ((ulong) DateTime.UtcNow.Ticks) * 100;
        }

        bool IsLittleEndian;

        ushort SwapBytes(ushort x)
        {
            if (!IsLittleEndian)
                return x;
            return (ushort)((ushort)((x & 0xff) << 8) | ((x >> 8) & 0xff));
        }

        public uint SwapBytes(uint x)
        {
            if (!IsLittleEndian)
                return x;
            return ((x & 0x000000ff) << 24) +
                   ((x & 0x0000ff00) << 8) +
                   ((x & 0x00ff0000) >> 8) +
                   ((x & 0xff000000) >> 24);
        }

        public ulong SwapBytes(ulong x)
        {
            if (!IsLittleEndian)
                return x;
            return
                  ((0x00000000000000ff) & (x >> 56)
                 | (0x000000000000ff00) & (x >> 40)
                 | (0x0000000000ff0000) & (x >> 24)
                 | (0x00000000ff000000) & (x >> 8)
                 | (0x000000ff00000000) & (x << 8)
                 | (0x0000ff0000000000) & (x << 24)
                 | (0x00ff000000000000) & (x << 40)
                 | (0xff00000000000000) & (x << 56));
        }
    }
}
